#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <malloc.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/un.h>
#include <sys/poll.h>

#include "defs.h"
#include "db.h"
#include "client.h"

#define IPC_SOCK_PATH 	"/tmp/gnomemsg_%s"

static char *hdr_string[] = { "",
	"MSG_LOGIN",
	"MSG_LOGOUT",
	"MSG_UINFO",
	"MSG_TINFO",
	"MSG_FINTASK",
	"MSG_TASK",
	"MSG_ERR",
	"MSG_ACK",
	"MSG_QUSER",
	"MSG_NOUSER",
	"MSG_NOLOGIN",
	"MSG_EOF",
	"MSG_STATUS",
	"MSG_CONTACT",
	"MSG_MESG",
	"MSG_NOTIFY",
	"MSG_CHGPASS",
	"MSG_RETRMSGS",
	"MSG_DELMSGS"
};

client_info client;

int auth_client(int sockfd, client_info * client)
{

	char hdr;
	msg_login login_msg;

	readin(sockfd, &login_msg, sizeof(msg_login));

	printf("%s ==> port: %d ntohs(port): %d\n", login_msg.login,
	       login_msg.port, ntohs(login_msg.port));

	if (register_user(&login_msg, client->ip) != 1) {
		hdr = MSG_ERR;
		writeout(sockfd, &hdr, sizeof(hdr));
		printf("%s ==> Error logging in.\n", login_msg.login);
		printf("%s ==> Thread dying!\n", login_msg.login);

		return -1;

	} else {
		hdr = MSG_ACK;
		writeout(sockfd, &hdr, sizeof(hdr));
		printf("%s ==> User %s logged in.\n", login_msg.login,
		       login_msg.login);

		strcpy(client->login, login_msg.login);
		client->status = login_msg.status;
		client->port = login_msg.port;

		return 1;

	}

}

void handle_client(int sockfd, struct sockaddr_in *addr)
{
	char *packet;
	char hdr;

	struct pollfd fds[2];

	int ipc_sock;
	int ipc_client;
	struct sockaddr_un ipc_addr;
	int len;

	msg_uinfo *uinfo_msg;
	msg_task *tinfo_msg;
	msg_quser *quser_msg;
	msg_mesg_dated *mesg_msg;
	msg_item *item;
	
	list_uinfo *uinfo_list;
	item_list *list;
	item_list *ptr;

	client_info contact;

	client.ip = addr->sin_addr.s_addr;

	readin(sockfd, &hdr, sizeof(hdr));

	if (hdr != MSG_LOGIN) {
		printf("luser ==> Bad Login packet from client.\n");
		printf("luser ==> Thread dying!\n");
		return;
	}

	if (auth_client(sockfd, &client) != 1) {
		return;
	}

	clear_notify(client.login);

	ipc_sock = socket(AF_UNIX, SOCK_STREAM, 0);
	ipc_addr.sun_family = AF_UNIX;
	sprintf(ipc_addr.sun_path, IPC_SOCK_PATH, client.login);
	unlink(ipc_addr.sun_path);

	if (bind(ipc_sock, (struct sockaddr *) &ipc_addr, sizeof(ipc_addr))
	    == -1) {
		printf("%s ==> couldn't bind to IPC socket.\n",
		       client.login);
		printf("%s ==> Thread dying!\n", client.login);
		return;
	}

	listen(ipc_sock, 10);

	printf("%s ==> opening socket %s for IPC\n", client.login,
	       ipc_addr.sun_path);

	fds[0].fd = ipc_sock;
	fds[0].events = POLLIN | POLLERR | POLLHUP;

	fds[1].fd = sockfd;
	fds[1].events = POLLIN | POLLERR | POLLHUP;

	while (client.status != STATUS_OFFLINE) {

		if (poll(fds, 2, -1) == -1)
			break;

		if (fds[1].revents & POLLIN) {

			readin(sockfd, &hdr, sizeof(hdr));
			printf("%s ==> %d (%s)\n", client.login, hdr,
			       hdr_string[hdr]);

			switch (hdr) {

			case MSG_STATUS:

				readin(sockfd, &hdr, sizeof(hdr));

				if (hdr >= STATUS_OFFLINE
				    && hdr <= STATUS_NA) {
					status_change(client.login, hdr);
					client.status = hdr;
					hdr = MSG_ACK;
					writeout(sockfd, &hdr,
						 sizeof(hdr));
					notify_clients(&client);
					break;
				} else {
					hdr = MSG_ERR;
					writeout(sockfd, &hdr,
						 sizeof(hdr));
					printf
					    ("%s ==> Bad Status.\n",
					     client.login);
					break;
				}


			case MSG_QUSER:
				packet = (char *) malloc(SIZE_ULEN);
				readin(sockfd, packet, SIZE_ULEN);

				quser_msg = (msg_quser *)
				    malloc(sizeof(msg_quser));

				if(get_ip(packet, quser_msg)==0) {
					hdr = MSG_QUSER;
					
					writeout(sockfd, &hdr, sizeof(hdr));
					writeout(sockfd, quser_msg,
					 sizeof(msg_quser));
				} else {
					hdr=MSG_NOUSER;
					writeout(sockfd, &hdr, sizeof(hdr));
				}

				free(packet);
				free(quser_msg);
				break;

			case MSG_UINFO:

				uinfo_msg = (msg_uinfo *)
				    malloc(sizeof(msg_uinfo));
				readin(sockfd, uinfo_msg,
				       sizeof(msg_uinfo));

				uinfo_list = get_uinfo(uinfo_msg);
				free(uinfo_msg);

				while (uinfo_list) {

					hdr = MSG_UINFO;
					writeout(sockfd, &hdr,
						 sizeof(hdr));
					writeout(sockfd,
						 &(uinfo_list->info),
						 sizeof(uinfo));
					uinfo_list = (list_uinfo *)
					    uinfo_list->next;
				}

				hdr = MSG_EOF;
				writeout(sockfd, &hdr, sizeof(hdr));

				break;

			case MSG_NOTIFY:

				clear_notify(client.login);

				readin(sockfd, &hdr, 1);
				len=hdr;

				hdr = MSG_ACK;
				
				packet = (char *) malloc(SIZE_ULEN);
				while (len--) {
					readin(sockfd, packet, SIZE_ULEN);

					if(add_notify(client.login, packet) != 0)
						hdr=MSG_NOUSER;
				}

				free(packet);

				writeout(sockfd, &hdr, sizeof(hdr));

				notify_clients(&client);

				break;

			case MSG_MESG:

				item=(msg_item *) malloc(sizeof(msg_item));
				item->type=MSG_MESG;

				mesg_msg=
				    (msg_mesg_dated *)
				    malloc(sizeof(msg_mesg_dated));

				readin(sockfd, &mesg_msg->hdr,
				       sizeof(msg_mesg_hdr));

				mesg_msg->mesg=
				    (char *) malloc(mesg_msg->hdr.length+1);
				readin(sockfd, mesg_msg->mesg, mesg_msg->hdr.length);
				mesg_msg->mesg[mesg_msg->hdr.length]=0;
				mesg_msg->hdr.date.hour=mesg_msg->hdr.date.min=mesg_msg->hdr.date.sec=mesg_msg->hdr.date.day=mesg_msg->hdr.date.month=mesg_msg->hdr.date.year=0;

				item->data=mesg_msg;

				hdr = MSG_ACK;
				
				if(store_mesg(item) !=0)
					hdr=MSG_NOUSER;
				
				free(mesg_msg->mesg);
				free(mesg_msg);

				writeout(sockfd, &hdr, sizeof(hdr));
				break;

			case MSG_RETRMSGS:

				list=get_offline_msgs();

				while(list) {

					item=&list->item;
					writeout(sockfd, &item->type, sizeof(item->type));

					mesg_msg=(msg_mesg_dated *) item->data;
					
					writeout(sockfd, &mesg_msg->hdr, sizeof(msg_mesg_hdr_dated));
					writeout(sockfd, mesg_msg->mesg, mesg_msg->hdr.length);

					free(mesg_msg->mesg);
					free(mesg_msg);

					ptr=list;
					list=(item_list *) list->next;
					free(ptr);
				}

				hdr=MSG_EOF;
				writeout(sockfd, &hdr, sizeof(hdr));

				readin(sockfd, &hdr, sizeof(hdr));
				if(hdr == MSG_ACK) del_offline_msgs();

				break;

			default:

				hdr = MSG_ERR;
				//writeout(sockfd, &hdr, sizeof(hdr));

			}
		} else if (fds[0].revents & POLLIN) {

			len = sizeof(ipc_addr);
			ipc_client =
			    accept(ipc_sock, (struct sockaddr *) &ipc_addr,
				   &len);
			readin(ipc_client, &contact, sizeof(contact));
			close(ipc_client);

			printf("%s ==> Message recieved from %s:\n",
			       client.login, contact.login);
			printf("%s ==> Status: %d\n", client.login,
			       contact.status);
			hdr = MSG_CONTACT;
			writeout(sockfd, &hdr, sizeof(hdr));
			writeout(sockfd, &contact, sizeof(contact));
		} else {
			status_change(client.login, STATUS_OFFLINE);
			clear_notify(client.login);
			close(sockfd);
			printf("%s ==> Thread dying!\n", client.login);
			return;
		}

	}

	close(ipc_sock);
	unlink(ipc_addr.sun_path);
	clear_notify(client.login);
	printf("%s ==> Thread dying!\n", client.login);
}

void readin(int sockfd, void *ptr, int len)
{
	int recv_len = recv(sockfd, ptr, len, MSG_NOSIGNAL | MSG_WAITALL);
	if (recv_len != len) {
		printf
		    ("%s ==> Illegal packet recieved from client. Disconnecting.\n",
		     client.login);
		status_change(client.login, STATUS_OFFLINE);
		client.status = STATUS_OFFLINE;
		notify_clients(&client);
		clear_notify(client.login);
		close(sockfd);
		exit(1);
	}
}

void writeout(int sockfd, void *ptr, int len)
{
	if (send(sockfd, ptr, len, MSG_NOSIGNAL) != len) {
		printf
		    ("%s ==> Illegal packet recieved from client. Disconnecting.\n",
		     client.login);
		status_change(client.login, STATUS_OFFLINE);
		client.status = STATUS_OFFLINE;
		notify_clients(&client);
		clear_notify(client.login);
		close(sockfd);
		exit(1);
	}
}

void notify_clients(client_info * client)
{
	list_notify *list;
	struct sockaddr_un ipc_addr;
	int sockfd;
	ipc_addr.sun_family = AF_UNIX;

	list = get_notify(client->login);

	while (list) {
		sprintf(ipc_addr.sun_path, IPC_SOCK_PATH,
			list->client.login);
		sockfd = socket(AF_UNIX, SOCK_STREAM, 0);

		printf("%s ==> Opening socket to %s\n", client->login,
		       ipc_addr.sun_path);

		if (connect
		    (sockfd, (struct sockaddr *) &ipc_addr,
		     sizeof(ipc_addr)) != -1) {

			printf("%s ==> Notifying %s thru %s\n",
			       client->login, list->client.login,
			       ipc_addr.sun_path);
			writeout(sockfd, client, sizeof(client_info));
			close(sockfd);
		} else {
			printf("%s ==> connection to %s failed\n",
			       client->login, ipc_addr.sun_path);
		}

		list = (list_notify *) list->next;
	}

}
