#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>

#include <signal.h>

#include "defs.h"
#include "client.h"
#include "db.h"

#define FORK

int server_sockfd, client_sockfd;

void quit(int sig)
{
	printf("Got signal %d!\n", sig);
	close(client_sockfd);
	close(server_sockfd);
	exit(1);
}

int main()
{
	socklen_t server_len, client_len;
	struct sockaddr_in server_addr, client_addr;
	server_sockfd = socket(AF_INET, SOCK_STREAM, 0);

	server_addr.sin_family = AF_INET;
	server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	server_addr.sin_port = htons(SERVER_PORT);
	server_len = sizeof(server_addr);

	signal(SIGINT, quit);
	signal(SIGSEGV, quit);

	if (bind
	    (server_sockfd, (struct sockaddr *) &server_addr,
	     server_len) != 0) {
		printf("Error Binding socket.\n");
		exit(1);
	}

	listen(server_sockfd, 5);

	signal(SIGCHLD, SIG_IGN);

	init_db();

	printf("Waiting for connections...\n");

	while (1) {

		printf("Starting New Thread!\n");

		client_len = sizeof(client_addr);
		client_sockfd = accept(server_sockfd,
				       (struct sockaddr *) &client_addr,
				       &client_len);

#ifdef FORK
		if (fork() == 0) {
#else
		if (1) {
#endif
			handle_client(client_sockfd, &client_addr);
			close(client_sockfd);
			exit(0);

		} else {
			close(client_sockfd);
		}

	}

}
